import 'package:adeptgraphql/adeptgraphql.dart';
import 'package:graphql_flutter/graphql_flutter.dart';
import 'package:http/http.dart' as http;
import 'package:adeptgraphql/src/api/graphql_api.dart' as GraphQLAPI;
import 'dart:async';

/// Created by krrigan on 04/28/20.

/// This class works as abstraction for GraphQL api services, allowing easy data parsing

abstract class GraphQLAPIService<I extends GraphQLService, DataType,
    ErrorType> {
  I service;
  DataType? dataType;
  ErrorType? errorType;
  GraphQLAPI.GraphQLAPI graphQLAPI;
  QueryResponseConverter<DataType> queryResponseConverter;

  GraphQLAPIService(
    this.service, {
    this.dataType,
    this.errorType,
    GraphQLAPI.GraphQLAPI? graphQLAPI,
    QueryResponseConverter<DataType>? queryResponseConverter,
  })  : graphQLAPI = graphQLAPI ?? GraphQLAPI.graphQLAPI,
        queryResponseConverter = queryResponseConverter ??
            QueryResponseConverter<DataType>(dataType: dataType);

  dynamic getAuthorization();

  updateHttpClient(http.Client httpClient) {
    graphQLAPI.init(httpClient: httpClient);
  }

  Future<QueryResult> getSaveResponse(
      Future<QueryResult> futureResponse) async {
    QueryResult response;
    try {
      response = await futureResponse;
    } catch (e) {
      if (e is QueryResult) {
        response = e;
      } else {
        rethrow;
      }
    }
    return response;
  }

  Future<QueryResponse<DataType, ErrorType>> parseResponse(
      Future<QueryResult> futureResponse,
      DataTypeQueryResponseConverter<DataType?>?
          dataTypeResponseConverter) async {
    return genericParseResponse(
        futureResponse,
        dataType,
        dataTypeResponseConverter ??
            queryResponseConverter.toDataResponseConverter);
  }

  Future<QueryResponse<List<DataType>, ErrorType>> parseResponseAsList(
      Future<QueryResult> futureResponse,
      DataTypeQueryResponseConverter<List<DataType>?>?
          dataTypeResponseConverter) async {
    return genericParseResponseAsList(
        futureResponse,
        dataType,
        dataTypeResponseConverter ??
            queryResponseConverter.toDataListResponseConverter);
  }

  Future<QueryResponse<GenericDataType, ErrorType>>
      genericParseResponse<GenericDataType>(
          Future<QueryResult> futureResponse,
          GenericDataType? dataType,
          DataTypeQueryResponseConverter<GenericDataType?>?
              dataTypeResponseConverter) async {
    QueryResult queryResult = await getSaveResponse(futureResponse);
    QueryResponse<GenericDataType, ErrorType> response;
    try {
      dataTypeResponseConverter ??=
          QueryResponseConverter<GenericDataType>(dataType: dataType)
              .toDataResponseConverter;
      response = QueryResponse.fromMapQueryResult<GenericDataType, ErrorType>(
          dataTypeResponseConverter: dataTypeResponseConverter,
          queryResult: queryResult,
          errorType: errorType);
    } catch (e) {
      response = QueryResponse<GenericDataType, ErrorType>(
        data: null,
        errorType: errorType,
        exception: queryResult.exception,
        source: queryResult.source,
        timestamp: queryResult.timestamp,
      );
      print(e);
    }
    return response;
  }

  Future<QueryResponse<List<GenericDataType>, ErrorType>>
      genericParseResponseAsList<GenericDataType>(
          Future<QueryResult> futureResponse,
          GenericDataType? dataType,
          DataTypeQueryResponseConverter<List<GenericDataType>?>?
              dataTypeResponseConverter) async {
    QueryResult queryResult = await getSaveResponse(futureResponse);
    QueryResponse<List<GenericDataType>, ErrorType> response;
    try {
      dataTypeResponseConverter ??=
          QueryResponseConverter<GenericDataType>(dataType: dataType)
              .toDataListResponseConverter;
      response = QueryResponse.fromListQueryResult<GenericDataType, ErrorType>(
          dataTypeResponseConverter: dataTypeResponseConverter,
          queryResult: queryResult,
          errorType: errorType);
    } catch (e) {
      response = QueryResponse<List<GenericDataType>, ErrorType>(
        data: null,
        errorType: errorType,
        exception: queryResult.exception,
        source: queryResult.source,
        timestamp: queryResult.timestamp,
      );
      print(e);
    }
    return response;
  }
}
