import 'package:adeptgraphql/adeptgraphql.dart';
import 'package:adeptutils/adeptutils.dart';
import 'package:graphql_flutter/graphql_flutter.dart';

/// Class used for GraphQL api responses

class QueryResponse<DataType, ErrorType> {
  GraphQLError? _mainGraphQLError;
  DateTime timestamp;
  List<ErrorType> _errors = [];
  ErrorType? _mainError;
  final ErrorType _errorType;

  /// The source of the result data.
  ///
  /// null when unexecuted.
  /// Will be set when encountering an error during any execution attempt
  QueryResultSource? source;
  DataType? data;
  OperationException? exception;

  QueryResponse({
    this.data,
    ErrorType? errorType,
    this.exception,
    bool? loading,
    bool? optimistic,
    QueryResultSource? source,
    DateTime? timestamp,
  })  : _errorType = errorType ??
            (GraphQLBaseError()
                as ErrorType), //TODO check if this breaks anything
        timestamp = timestamp ?? DateTime.now(),
        source = source ??
            ((loading == true)
                ? QueryResultSource.loading
                : (optimistic == true)
                    ? QueryResultSource.optimisticResult
                    : null);

  List<ErrorType> get errors {
    if (_errors.isEmpty && hasException) {
      if ((exception?.graphqlErrors.isNotEmpty ?? false)) {
        _errors = [];
        try {
          for (var graphQLError in exception!.graphqlErrors) {
            if (_errorType is Jsonable) {
              Map<String, dynamic> graphQLErrorMap = {
                'message': graphQLError.message,
                'locations': graphQLError.locations,
                'path': graphQLError.path,
                'extensions': graphQLError.extensions,
              };
              ErrorType? parsedError = (_errorType as Jsonable)
                  .fromJsonMap(graphQLErrorMap) as ErrorType?;
              if (parsedError != null) _errors.add(parsedError);
            }
            // else _errors.add(graphQLError.extensions as ErrorType?);
          }
        } catch (e) {
          print(e);
        }
      }
    }
    return _errors;
  }

  ErrorType? get mainError {
    if (_mainError == null && errors.isNotEmpty) _mainError = errors[0];
    return _mainError;
  }

  GraphQLError? get mainGraphQLError {
    if (_mainGraphQLError == null) {
      if ((exception?.graphqlErrors.isNotEmpty ?? false)) {
        _mainGraphQLError = exception!.graphqlErrors[0];
      }
    }
    return _mainGraphQLError;
  }

  String? get errorMessage {
    return mainGraphQLError?.message;
  }

  String? get errorCode {
    try {
      if (mainGraphQLError?.extensions?.isNotEmpty ?? false) {
        return mainGraphQLError!.extensions!['code'];
      }
    } catch (e) {
      print(e);
    }
    return null;
  }

  int? get errorStatusCode {
    try {
      if (mainGraphQLError?.extensions?.isNotEmpty ?? false) {
        Map<String, dynamic>? exception =
            mainGraphQLError!.extensions!['exception'];
        if ((exception?.isNotEmpty ?? false)) return exception!['status'];
      }
    } catch (e) {
      print(e);
    }
    return null;
  }

  /// Whether data has been specified from either the cache or network)
  bool get loading => source == QueryResultSource.loading;

  /// Whether an optimistic result has been specified
  ///   (may include eager results from the cache)
  bool get optimistic => source == QueryResultSource.optimisticResult;

  /// Whether the response includes an exception
  bool get hasException => exception != null;

  /// Whether the response includes an exception
  bool get isSuccessful => exception == null;

//  static QueryResponse fromQueryResult(
//      QueryResult queryResult, DataTypeResponseConverter dataTypeResponseConverter){
//    var data;
//    try{
//      data = dataTypeResponseConverter(queryResult);
//    }catch(e){
//      print(e);
//    }
//    return QueryResponse(
//      data: data,
//      exception: queryResult?.exception,
//      source: queryResult?.source,
//      timestamp: queryResult.timestamp,
//    );
//  }
//
//  static QueryResponse fromQueryResultOld<DataType extends Jsonable>(
//      QueryResult queryResult, DataType dataType,
//      DataTypeResponseConverter dataTypeResponseConverter){
//    DataType data;
//    try{
//      if(queryResult?.data != null)
//        return queryResult?.data is Map
//            ? fromMapQueryResult(queryResult, dataType, dataTypeResponseConverter)
//            : fromListQueryResult(queryResult, dataType, dataTypeResponseConverter);
//    }catch(e){
//      print(e);
//    }
//    return QueryResponse<DataType>(
//      data: data,
//      exception: queryResult?.exception,
//      source: queryResult?.source,
//      timestamp: queryResult.timestamp,
//    );
//  }

  static QueryResponse<DataType, ErrorType>
      fromMapQueryResult<DataType, ErrorType>({
    required DataTypeQueryResponseConverter<DataType?>
        dataTypeResponseConverter,
    QueryResult? queryResult,
    ErrorType? errorType,
  }) {
    DataType? dataType;
    try {
      dataType = dataTypeResponseConverter(queryResult);
    } catch (e) {
      print(e);
    }
    return QueryResponse<DataType, ErrorType>(
      data: dataType,
      errorType: errorType,
      exception: queryResult?.exception,
      source: queryResult?.source,
      timestamp: queryResult?.timestamp,
    );
  }

  static QueryResponse<List<DataType>, ErrorType>
      fromListQueryResult<DataType, ErrorType>({
    required DataTypeQueryResponseConverter<List<DataType>?>
        dataTypeResponseConverter,
    QueryResult? queryResult,
    ErrorType? errorType,
  }) {
    List<DataType>? dataTypeList;
    try {
      dataTypeList = dataTypeResponseConverter(queryResult);
    } catch (e) {
      print(e);
    }
    return QueryResponse<List<DataType>, ErrorType>(
      data: dataTypeList,
      errorType: errorType,
      exception: queryResult?.exception,
      source: queryResult?.source,
      timestamp: queryResult?.timestamp,
    );
  }
}
