import 'package:adeptarch_base/adeptarch_base.dart';
import 'package:adeptgraphql/src/api/query_response.dart';
import 'package:adeptgraphql/src/api/query_response_conveter.dart';
import 'package:adeptgraphql/src/resource/subscription_resource.dart';
import 'package:adeptutils/adeptutils.dart';
import 'dart:async';

import 'package:graphql_flutter/graphql_flutter.dart';

/// A generic class that can provide a subscription resource

typedef SubscriptionCall = Future<Stream<QueryResult>> Function();
typedef OnSubscriptionDataSuccess<DataType, ErrorType>
    = Future<SubscriptionResource<DataType>> Function(
        {SubscriptionResource<DataType>? resource,
        QueryResponse<DataType, ErrorType>? response});
typedef OnSubscriptionDataError<DataType, ErrorType>
    = Future<SubscriptionResource<DataType>> Function(
        {dynamic exception, QueryResponse<DataType, ErrorType>? response});

typedef OnSubscriptionFailed<DataType> = Future<SubscriptionResource<DataType>>
    Function(dynamic exception);
typedef OnSubscriptionErrorManagement<DataType, ErrorType>
    = Future<SubscriptionResource<DataType>> Function(
        SubscriptionResource<DataType>? resource,
        {QueryResponse<DataType, ErrorType>? response});

class SubscriptionResourceNotifier<DataType, ErrorType> {
  StreamSubscription? _streamSubscription;
  final AdeptPublishSubject<SubscriptionResource<DataType?>> _publishSubject;
  SubscriptionCall subscriptionCall;
  OnSubscriptionDataSuccess<DataType?, ErrorType?>? onSubscriptionDataSuccess;
  OnSubscriptionDataError<DataType?, ErrorType?>? onSubscriptionDataError;
  OnSubscriptionFailed<DataType?>? onSubscriptionFailed;
  OnSubscriptionErrorManagement<DataType?, ErrorType?>?
      onSubscriptionErrorManagement;
  final QueryResponseConverter<DataType?> _dataTypeResponseConverter;

  SubscriptionResourceNotifier(
      {DataType? dataType,
      AdeptPublishSubject<SubscriptionResource<DataType?>>? publishSubject,
      required this.subscriptionCall,
      this.onSubscriptionDataSuccess,
      this.onSubscriptionDataError,
      this.onSubscriptionFailed,
      this.onSubscriptionErrorManagement,
      QueryResponseConverter<DataType?>? dataTypeResponseConverter})
      : _dataTypeResponseConverter = dataTypeResponseConverter ??
            QueryResponseConverter<DataType>(dataType: dataType),
        _publishSubject = publishSubject ??
            AdeptPublishSubject<SubscriptionResource<DataType?>>();

  AdeptPublishSubject<SubscriptionResource<DataType?>> get publishSubject =>
      _publishSubject;

  Future<SubscriptionResource<DataType?>> handleOnSubscriptionDataSuccess(
      {SubscriptionResource<DataType?>? resource,
      QueryResponse<DataType?, ErrorType?>? response}) async {
    if (onSubscriptionDataSuccess != null) {
      resource = await onSubscriptionDataSuccess!(
          resource: resource, response: response);
    } else {
      if (response != null && response.isSuccessful) {
        return handleOnSubscriptionDataSuccess(
            resource: SubscriptionResource.data(response.data,
                extras: response.errors));
      }
    }
    resource ??= SubscriptionResource<DataType>.error(
        null, 'Empty GraphQL Subscription data response');
    publishSubject.post(resource);
    return resource;
  }

  Future<SubscriptionResource<DataType?>> handleOnSubscriptionDataError(
      {dynamic exception,
      QueryResponse<DataType?, ErrorType?>? response}) async {
    SubscriptionResource<DataType?> resource;
    if (onSubscriptionDataError != null) {
      resource = await onSubscriptionDataError!(
          exception: exception, response: response);
    } else {
      resource = SubscriptionResource.error(
          null, ExceptionUtils.parseException(exception).toString(),
          exception: exception,
          extras: response?.errors,
          code: response?.errorStatusCode);
      if (onSubscriptionErrorManagement != null) {
        resource =
            await onSubscriptionErrorManagement!(resource, response: response);
      }
    }
    publishSubject.post(resource);
    return resource;
  }

  Future<SubscriptionResource<DataType?>> handleOnSubscriptionFailed(
      dynamic exception) async {
    SubscriptionResource<DataType?> resource;
    if (onSubscriptionFailed != null) {
      resource = await onSubscriptionFailed!(exception);
    } else {
      resource = SubscriptionResource.error(null, exception?.toString(),
          exception: exception);
      if (onSubscriptionErrorManagement != null) {
        resource = await onSubscriptionErrorManagement!(resource);
      }
    }
    publishSubject.post(resource);
    return resource;
  }

  void _onData(final QueryResult data) async {
    try {
      QueryResponse<DataType, ErrorType> response =
          QueryResponse.fromMapQueryResult(
        dataTypeResponseConverter:
            _dataTypeResponseConverter.toDataResponseConverter,
        queryResult: data,
      );

      if (response.isSuccessful) {
        await handleOnSubscriptionDataSuccess(response: response);
      } else {
        await handleOnSubscriptionDataError(response: response);
      }
    } catch (e) {
      await handleOnSubscriptionDataError(exception: e);
    }
  }

  void _onError(final Object error) {
    handleOnSubscriptionDataError(
        exception: (error is SubscriptionError) ? error.payload : error);
  }

  void _onDone() => publishSubject.post(SubscriptionResource<DataType>.done());

  AdeptPublishSubject<SubscriptionResource<DataType?>> asPublishSubject() {
    start();
    return publishSubject;
  }

  Future<void> start() async {
    try {
      _streamSubscription = (await subscriptionCall()).listen(
        _onData,
        onError: _onError,
        onDone: _onDone,
      );
    } catch (e) {
      handleOnSubscriptionFailed(e);
    }
  }

  void pause([Future<void>? resumeSignal]) =>
      _streamSubscription?.pause(resumeSignal);
  void resume() => _streamSubscription?.resume();
  Future<void> cancel() async {
    if (_streamSubscription != null) return _streamSubscription!.cancel();
  }

  bool get isPaused => _streamSubscription?.isPaused ?? false;
}
